import { warning as b, invariant as S } from "../utils/errors.mjs";
import { createBox as P } from "../projection/geometry/models.mjs";
import { isRefObject as T } from "../utils/is-ref-object.mjs";
import { initPrefersReducedMotion as F } from "../utils/reduced-motion/index.mjs";
import { hasReducedMotionListener as y, prefersReducedMotion as R } from "../utils/reduced-motion/state.mjs";
import { SubscriptionManager as j } from "../utils/subscription-manager.mjs";
import { motionValue as w } from "../value/index.mjs";
import { isWillChangeMotionValue as N } from "../value/use-will-change/is.mjs";
import { isMotionValue as c } from "../value/utils/is-motion-value.mjs";
import { transformProps as E } from "./html/utils/transform.mjs";
import { isControllingVariants as x, isVariantNode as B } from "./utils/is-controlling-variants.mjs";
import { isVariantLabel as L } from "./utils/is-variant-label.mjs";
import { updateMotionValuesFromProps as I } from "./utils/motion-values.mjs";
import { resolveVariantFromProps as O } from "./utils/resolve-variants.mjs";
import { warnOnce as A } from "../utils/warn-once.mjs";
import { featureDefinitions as g } from "../motion/features/definitions.mjs";
import { variantProps as M } from "./utils/variant-props.mjs";
import { visualElementStore as f } from "./store.mjs";
import { frame as m, cancelFrame as V } from "../frameloop/frame.mjs";
const C = Object.keys(g), U = C.length, v = [
  "AnimationStart",
  "AnimationComplete",
  "Update",
  "BeforeLayoutMeasure",
  "LayoutMeasure",
  "LayoutAnimationStart",
  "LayoutAnimationComplete"
], _ = M.length;
class ot {
  constructor({ parent: t, props: e, presenceContext: i, reducedMotionConfig: s, visualState: r }, u = {}) {
    this.current = null, this.children = /* @__PURE__ */ new Set(), this.isVariantNode = !1, this.isControllingVariants = !1, this.shouldReduceMotion = null, this.values = /* @__PURE__ */ new Map(), this.features = {}, this.valueSubscriptions = /* @__PURE__ */ new Map(), this.prevMotionValues = {}, this.events = {}, this.propEventSubscriptions = {}, this.notifyUpdate = () => this.notify("Update", this.latestValues), this.render = () => {
      this.current && (this.triggerBuild(), this.renderInstance(this.current, this.renderState, this.props.style, this.projection));
    }, this.scheduleRender = () => m.render(this.render, !1, !0);
    const { latestValues: h, renderState: n } = r;
    this.latestValues = h, this.baseTarget = { ...h }, this.initialValues = e.initial ? { ...h } : {}, this.renderState = n, this.parent = t, this.props = e, this.presenceContext = i, this.depth = t ? t.depth + 1 : 0, this.reducedMotionConfig = s, this.options = u, this.isControllingVariants = x(e), this.isVariantNode = B(e), this.isVariantNode && (this.variantChildren = /* @__PURE__ */ new Set()), this.manuallyAnimateOnMount = !!(t && t.current);
    const { willChange: o, ...p } = this.scrapeMotionValuesFromProps(e, {});
    for (const a in p) {
      const l = p[a];
      h[a] !== void 0 && c(l) && (l.set(h[a], !1), N(o) && o.add(a));
    }
  }
  /**
   * This method takes React props and returns found MotionValues. For example, HTML
   * MotionValues will be found within the style prop, whereas for Three.js within attribute arrays.
   *
   * This isn't an abstract method as it needs calling in the constructor, but it is
   * intended to be one.
   */
  scrapeMotionValuesFromProps(t, e) {
    return {};
  }
  mount(t) {
    this.current = t, f.set(t, this), this.projection && !this.projection.instance && this.projection.mount(t), this.parent && this.isVariantNode && !this.isControllingVariants && (this.removeFromVariantTree = this.parent.addVariantChild(this)), this.values.forEach((e, i) => this.bindToMotionValue(i, e)), y.current || F(), this.shouldReduceMotion = this.reducedMotionConfig === "never" ? !1 : this.reducedMotionConfig === "always" ? !0 : R.current, process.env.NODE_ENV !== "production" && A(this.shouldReduceMotion !== !0, "You have Reduced Motion enabled on your device. Animations may not appear as expected."), this.parent && this.parent.children.add(this), this.update(this.props, this.presenceContext);
  }
  unmount() {
    f.delete(this.current), this.projection && this.projection.unmount(), V(this.notifyUpdate), V(this.render), this.valueSubscriptions.forEach((t) => t()), this.removeFromVariantTree && this.removeFromVariantTree(), this.parent && this.parent.children.delete(this);
    for (const t in this.events)
      this.events[t].clear();
    for (const t in this.features)
      this.features[t].unmount();
    this.current = null;
  }
  bindToMotionValue(t, e) {
    const i = E.has(t), s = e.on("change", (u) => {
      this.latestValues[t] = u, this.props.onUpdate && m.update(this.notifyUpdate, !1, !0), i && this.projection && (this.projection.isTransformDirty = !0);
    }), r = e.on("renderRequest", this.scheduleRender);
    this.valueSubscriptions.set(t, () => {
      s(), r();
    });
  }
  sortNodePosition(t) {
    return !this.current || !this.sortInstanceNodePosition || this.type !== t.type ? 0 : this.sortInstanceNodePosition(this.current, t.current);
  }
  loadFeatures({ children: t, ...e }, i, s, r) {
    let u, h;
    if (process.env.NODE_ENV !== "production" && s && i) {
      const n = "You have rendered a `motion` component within a `LazyMotion` component. This will break tree shaking. Import and render a `m` component instead.";
      e.ignoreStrict ? b(!1, n) : S(!1, n);
    }
    for (let n = 0; n < U; n++) {
      const o = C[n], { isEnabled: p, Feature: a, ProjectionNode: l, MeasureLayout: d } = g[o];
      l && (u = l), p(e) && (!this.features[o] && a && (this.features[o] = new a(this)), d && (h = d));
    }
    if (!this.projection && u) {
      this.projection = new u(this.latestValues, this.parent && this.parent.projection);
      const { layoutId: n, layout: o, drag: p, dragConstraints: a, layoutScroll: l, layoutRoot: d } = e;
      this.projection.setOptions({
        layoutId: n,
        layout: o,
        alwaysMeasureLayout: !!p || a && T(a),
        visualElement: this,
        scheduleRender: () => this.scheduleRender(),
        /**
         * TODO: Update options in an effect. This could be tricky as it'll be too late
         * to update by the time layout animations run.
         * We also need to fix this safeToRemove by linking it up to the one returned by usePresence,
         * ensuring it gets called if there's no potential layout animations.
         *
         */
        animationType: typeof o == "string" ? o : "both",
        initialPromotionConfig: r,
        layoutScroll: l,
        layoutRoot: d
      });
    }
    return h;
  }
  updateFeatures() {
    for (const t in this.features) {
      const e = this.features[t];
      e.isMounted ? e.update() : (e.mount(), e.isMounted = !0);
    }
  }
  triggerBuild() {
    this.build(this.renderState, this.latestValues, this.options, this.props);
  }
  /**
   * Measure the current viewport box with or without transforms.
   * Only measures axis-aligned boxes, rotate and skew must be manually
   * removed with a re-render to work.
   */
  measureViewportBox() {
    return this.current ? this.measureInstanceViewportBox(this.current, this.props) : P();
  }
  getStaticValue(t) {
    return this.latestValues[t];
  }
  setStaticValue(t, e) {
    this.latestValues[t] = e;
  }
  /**
   * Make a target animatable by Popmotion. For instance, if we're
   * trying to animate width from 100px to 100vw we need to measure 100vw
   * in pixels to determine what we really need to animate to. This is also
   * pluggable to support Framer's custom value types like Color,
   * and CSS variables.
   */
  makeTargetAnimatable(t, e = !0) {
    return this.makeTargetAnimatableFromInstance(t, this.props, e);
  }
  /**
   * Update the provided props. Ensure any newly-added motion values are
   * added to our map, old ones removed, and listeners updated.
   */
  update(t, e) {
    (t.transformTemplate || this.props.transformTemplate) && this.scheduleRender(), this.prevProps = this.props, this.props = t, this.prevPresenceContext = this.presenceContext, this.presenceContext = e;
    for (let i = 0; i < v.length; i++) {
      const s = v[i];
      this.propEventSubscriptions[s] && (this.propEventSubscriptions[s](), delete this.propEventSubscriptions[s]);
      const r = t["on" + s];
      r && (this.propEventSubscriptions[s] = this.on(s, r));
    }
    this.prevMotionValues = I(this, this.scrapeMotionValuesFromProps(t, this.prevProps), this.prevMotionValues), this.handleChildMotionValue && this.handleChildMotionValue();
  }
  getProps() {
    return this.props;
  }
  /**
   * Returns the variant definition with a given name.
   */
  getVariant(t) {
    return this.props.variants ? this.props.variants[t] : void 0;
  }
  /**
   * Returns the defined default transition on this component.
   */
  getDefaultTransition() {
    return this.props.transition;
  }
  getTransformPagePoint() {
    return this.props.transformPagePoint;
  }
  getClosestVariantNode() {
    return this.isVariantNode ? this : this.parent ? this.parent.getClosestVariantNode() : void 0;
  }
  getVariantContext(t = !1) {
    if (t)
      return this.parent ? this.parent.getVariantContext() : void 0;
    if (!this.isControllingVariants) {
      const i = this.parent ? this.parent.getVariantContext() || {} : {};
      return this.props.initial !== void 0 && (i.initial = this.props.initial), i;
    }
    const e = {};
    for (let i = 0; i < _; i++) {
      const s = M[i], r = this.props[s];
      (L(r) || r === !1) && (e[s] = r);
    }
    return e;
  }
  /**
   * Add a child visual element to our set of children.
   */
  addVariantChild(t) {
    const e = this.getClosestVariantNode();
    if (e)
      return e.variantChildren && e.variantChildren.add(t), () => e.variantChildren.delete(t);
  }
  /**
   * Add a motion value and bind it to this visual element.
   */
  addValue(t, e) {
    e !== this.values.get(t) && (this.removeValue(t), this.bindToMotionValue(t, e)), this.values.set(t, e), this.latestValues[t] = e.get();
  }
  /**
   * Remove a motion value and unbind any active subscriptions.
   */
  removeValue(t) {
    this.values.delete(t);
    const e = this.valueSubscriptions.get(t);
    e && (e(), this.valueSubscriptions.delete(t)), delete this.latestValues[t], this.removeValueFromRenderState(t, this.renderState);
  }
  /**
   * Check whether we have a motion value for this key
   */
  hasValue(t) {
    return this.values.has(t);
  }
  getValue(t, e) {
    if (this.props.values && this.props.values[t])
      return this.props.values[t];
    let i = this.values.get(t);
    return i === void 0 && e !== void 0 && (i = w(e, { owner: this }), this.addValue(t, i)), i;
  }
  /**
   * If we're trying to animate to a previously unencountered value,
   * we need to check for it in our state and as a last resort read it
   * directly from the instance (which might have performance implications).
   */
  readValue(t) {
    var e;
    return this.latestValues[t] !== void 0 || !this.current ? this.latestValues[t] : (e = this.getBaseTargetFromProps(this.props, t)) !== null && e !== void 0 ? e : this.readValueFromInstance(this.current, t, this.options);
  }
  /**
   * Set the base target to later animate back to. This is currently
   * only hydrated on creation and when we first read a value.
   */
  setBaseTarget(t, e) {
    this.baseTarget[t] = e;
  }
  /**
   * Find the base target for a value thats been removed from all animation
   * props.
   */
  getBaseTarget(t) {
    var e;
    const { initial: i } = this.props, s = typeof i == "string" || typeof i == "object" ? (e = O(this.props, i)) === null || e === void 0 ? void 0 : e[t] : void 0;
    if (i && s !== void 0)
      return s;
    const r = this.getBaseTargetFromProps(this.props, t);
    return r !== void 0 && !c(r) ? r : this.initialValues[t] !== void 0 && s === void 0 ? void 0 : this.baseTarget[t];
  }
  on(t, e) {
    return this.events[t] || (this.events[t] = new j()), this.events[t].add(e);
  }
  notify(t, ...e) {
    this.events[t] && this.events[t].notify(...e);
  }
}
export {
  ot as VisualElement
};
